module dtwofish;

import twofish_import;
import std.exception;
import std.zlib;


enum DEFAULT_KEYSIZE = 256;
enum DEFAULT_ENCRYPTION_MODE = MODE_CBC;
// maked by mt19937
enum INIT_IV_TABLE = [0x7302c952, 0x3a8ff7a6, 0x6ca58c46, 0xc700379f];


private struct CipherText {
    size_t cipherText_size;
    size_t plainText_size;
    int use_deflate;
    ubyte cipherText_top;
}

class Twofish {
    keyInstance ki;
    cipherInstance ci;
    
    this() {
        init();
    }
    void init() {
        enforce(makeKey(&ki, DIR_ENCRYPT, DEFAULT_KEYSIZE, null) == TRUE);
        enforce(cipherInit(&ci, DEFAULT_ENCRYPTION_MODE, null) == TRUE);
    }

    void setKey(uint[8] key) {
        ki.key32[] = key.dup;
        reKey(&ki);
    }
    
    ubyte[] encrypt(ubyte[] plainText) {
        ci.iv32[] = INIT_IV_TABLE.dup;
        auto cpt = cast(const(ubyte)[])compress(plainText);
        int use_deflate = 0;
        if (cast(real)cpt.length / cast(real)plainText.length < 0.95) {
            plainText = cpt.dup;
            use_deflate = 1;
        }
        auto output = new ubyte[plainText.length + (CipherText.sizeof*2 - 1) +
                                (4 - plainText.length%4)];
        auto ct = cast(CipherText*)output.ptr;
        ct.use_deflate = use_deflate;
        enforce(blockEncrypt(&ci, &ki,
                             plainText.ptr,
                             plainText.length*8,
                             &ct.cipherText_top)
                  == plainText.length*8,
                "blockEncrypt error in Twofish.encrypt");
        ct.plainText_size  = plainText.length;
        ct.cipherText_size = output.length;
        return output;
    }
    ubyte[] decrypt(ubyte[] cipherText) {
        ci.iv32[] = INIT_IV_TABLE.dup;
        auto ct = cast(CipherText*)cipherText.ptr;
        auto output = new ubyte[ct.cipherText_size];
        enforce(blockDecrypt(&ci, &ki,
                             &ct.cipherText_top,
                             ct.cipherText_size*8,
                             output.ptr) == ct.cipherText_size*8,
                "blockDecrypt error in Twofish.decrypt");
        if (ct.use_deflate) {
            auto cpt = output[0 .. ct.plainText_size];
            return (cast(const(ubyte)[])uncompress(cpt)).dup;
        }
        else {
            return output[0 .. ct.plainText_size];
        }
    }
}

