/* Converted to D from aes.h by htod */
module twofish_import;
/* aes.h */

/* ---------- See examples at end of this file for typical usage -------- */

/* AES Cipher header file for ANSI C Submissions
	Lawrence E. Bassham III
	Computer Security Division
	National Institute of Standards and Technology

	This sample is to assist implementers developing to the
Cryptographic API Profile for AES Candidate Algorithm Submissions.
Please consult this document as a cross-reference.
	
	ANY CHANGES, WHERE APPROPRIATE, TO INFORMATION PROVIDED IN THIS FILE
MUST BE DOCUMENTED. CHANGES ARE ONLY APPROPRIATE WHERE SPECIFIED WITH
THE STRING "CHANGE POSSIBLE". FUNCTION CALLS AND THEIR PARAMETERS
CANNOT BE CHANGED. STRUCTURES CAN BE ALTERED TO ALLOW IMPLEMENTERS TO
INCLUDE IMPLEMENTATION SPECIFIC INFORMATION.
*/

/* Includes:
	Standard include files
*/

import std.c.stdio;

/*	Defines:
		Add any additional defines you need
*/

enum DIR_ENCRYPT = 0;
enum DIR_DECRYPT = 1;

enum ubyte MODE_ECB = 1;
enum ubyte MODE_CBC = 2;
enum ubyte  MODE_CFB1 = 3;

enum TRUE = 1;
enum FALSE = 0;

enum BAD_KEY_DIR = -1;
enum BAD_KEY_MAT = -2;
enum BAD_KEY_INSTANCE = -3;
enum BAD_CIPHER_MODE = -4;

enum BAD_CIPHER_STATE = -5;
/* CHANGE POSSIBLE: inclusion of algorithm specific defines */
/* TWOFISH specific definitions */
enum MAX_KEY_SIZE = 64;
enum MAX_IV_SIZE = 16;
enum BAD_INPUT_LEN = -6;
enum BAD_PARAMS = -7;
enum BAD_IV_MAT = -8;
enum BAD_ENDIAN = -9;

enum BAD_ALIGN32 = -10;
enum BLOCK_SIZE = 128;
enum MAX_ROUNDS = 16;
enum ROUNDS_128 = 16;
enum ROUNDS_192 = 16;
enum ROUNDS_256 = 16;
enum MAX_KEY_BITS = 256;
enum MIN_KEY_BITS = 128;
enum VALID_SIG = 0x48534946;
enum MCT_OUTER = 400;
enum MCT_INNER = 10000;

enum REENTRANT = 1;
enum INPUT_WHITEN = 0;
enum OUTPUT_WHITEN = (INPUT_WHITEN + BLOCK_SIZE/32);
enum ROUND_SUBKEYS = (OUTPUT_WHITEN + BLOCK_SIZE/32);
enum TOTAL_SUBKEYS = (ROUND_SUBKEYS + 2*MAX_ROUNDS);

/* Typedefs:
	Typedef'ed data storage elements. Add any algorithm specific
	parameters at the bottom of the structs as appropriate.
*/


alias uint[256][4] fullSbox;

/* The structure for key information */
//C     typedef struct 
//C     	{
//C     	BYTE direction;					/* Key used for encrypting or decrypting? */
//C     #if ALIGN32
//C     	BYTE dummyAlign[3];				/* keep 32-bit alignment */
//C     #endif
//C     	int  keyLen;					/* Length of the key */
//C     	char keyMaterial[MAX_KEY_SIZE+4];/* Raw key data in ASCII */

	/* Twofish-specific parameters: */
//C     	DWORD keySig;					/* set to VALID_SIG by makeKey() */
//C     	int	  numRounds;				/* number of rounds in cipher */
//C     	DWORD key32[MAX_KEY_BITS/32];	/* actual key bits, in dwords */
//C     	DWORD sboxKeys[MAX_KEY_BITS/64];/* key bits used for S-boxes */
//C     	DWORD subKeys[TOTAL_SUBKEYS];	/* round subkeys, input/output whitening bits */
//C     #if REENTRANT
//C     	fullSbox sBox8x32;				/* fully expanded S-box */
//C       #if defined(COMPILE_KEY) && defined(USE_ASM)
//C     #undef	VALID_SIG
//C     #define	VALID_SIG	 0x504D4F43		/* 'COMP':  C is compiled with -DCOMPILE_KEY */
//C     	DWORD cSig1;					/* set after first "compile" (zero at "init") */
//C     	void *encryptFuncPtr;			/* ptr to asm encrypt function */
//C     	void *decryptFuncPtr;			/* ptr to asm decrypt function */
//C     	DWORD codeSize;					/* size of compiledCode */
//C     	DWORD cSig2;					/* set after first "compile" */
//C     	BYTE  compiledCode[5000];		/* make room for the code itself */
//C       #endif
//C     #endif
//C     	} keyInstance;
struct keyInstance {
    ubyte direction;
    int keyLen;
    char[68] keyMaterial;
    uint keySig;
    int numRounds;
    uint[8] key32;
    uint[4] sboxKeys;
    uint[40] subKeys;
    fullSbox sBox8x32;
}
unittest {
    assert(keyInstance.sizeof == 4388);
}

/* The structure for cipher information */
//C     typedef struct 
//C     	{
//C     	BYTE  mode;						/* MODE_ECB, MODE_CBC, or MODE_CFB1 */
//C     #if ALIGN32
//C     	BYTE dummyAlign[3];				/* keep 32-bit alignment */
//C     #endif
//C     	BYTE  IV[MAX_IV_SIZE];			/* CFB1 iv bytes  (CBC uses iv32) */

	/* Twofish-specific parameters: */
//C     	DWORD cipherSig;				/* set to VALID_SIG by cipherInit() */
//C     	DWORD iv32[BLOCK_SIZE/32];		/* CBC IV bytes arranged as dwords */
//C     	} cipherInstance;
struct cipherInstance
{
    ubyte mode;
    ubyte[16] IV;
    uint cipherSig;
    uint[4] iv32;
}
unittest {
    assert(cipherInstance.sizeof == 40);
}


/* Function protoypes */
extern (C) {
    int  makeKey(keyInstance* key, ubyte direction, int keyLen, ubyte* keyMaterial);
    int  cipherInit(cipherInstance* cipher, ubyte mode, ubyte* IV);
    int  blockEncrypt(cipherInstance* cipher, keyInstance* key, ubyte* input, int inputLen, ubyte* outBuffer);
    int  blockDecrypt(cipherInstance* cipher, keyInstance* key, ubyte* input, int inputLen, ubyte* outBuffer);
    int  reKey(keyInstance *key);
}

/* API to check table usage, for use in ECB_TBL KAT */
enum TAB_DISABLE = 0;
enum TAB_ENABLE = 1;
enum TAB_RESET = 2;
enum TAB_QUERY = 3;
enum TAB_MIN_QUERY = 50;
int  TableOp(int op);

//C     #define		BlockCopy(d,s)	{ memcpy(d,s,BLOCK_SIZE/8); }


version(unittest) {
    import std.c.stdlib;
    import std.c.string;
    import std.string;

    /* max # blocks per call in TestTwofish */
    enum MAX_BLK_CNT = 4;
    
    /* keySize must be 128, 192, or 256 */
    int TestTwofish(ubyte mode, int keySize) {
        /* return 0 iff test passes */
        keyInstance    ki; /* key information, including tables */
        cipherInstance ci; /* keeps mode (ECB, CBC) and IV */
        ubyte[MAX_BLK_CNT*(BLOCK_SIZE/8)] plainText;
        ubyte[MAX_BLK_CNT*(BLOCK_SIZE/8)] cipherText;
        ubyte[MAX_BLK_CNT*(BLOCK_SIZE/8)] decryptOut;
        ubyte[BLOCK_SIZE/8] iv;
        int i, byteCnt;

        if (makeKey(&ki, DIR_ENCRYPT, keySize, null) != TRUE)
            return 1; /* 'dummy' setup for a 128-bit key */
        if (cipherInit(&ci, mode, null) != TRUE)
            return 1; /* 'dummy' setup for cipher */

        for (i=0; i<keySize/32; i++) /* select key bits */
            ki.key32[i]=0x10003 * rand();
        reKey(&ki); /* run the key schedule */

        if (mode != MODE_ECB) { /* set up random iv (if needed)*/
            for (i=0; i<iv.sizeof; i++)
                iv[i] = cast(ubyte)rand();
            memcpy(ci.iv32.ptr, iv.ptr, ci.iv32.sizeof);	/* copy the IV to ci */
        }

        /* select number of bytes to encrypt (multiple of block) */
        /* e.g., byteCnt = 16, 32, 48, 64 */
        byteCnt = (BLOCK_SIZE/8) * (1 + (rand() % MAX_BLK_CNT));

        for (i=0; i<byteCnt; i++) /* generate test data */
            plainText[i] = cast(ubyte)rand();

        /* encrypt the bytes */
        if (blockEncrypt(&ci, &ki, plainText.ptr,
                         byteCnt*8, cipherText.ptr) != byteCnt*8)
            return 1;

        /* decrypt the bytes */
        if (mode != MODE_ECB) /* first re-init the IV (if needed) */
            memcpy(ci.iv32.ptr, iv.ptr, ci.iv32.sizeof);

        if (blockDecrypt(&ci, &ki, cipherText.ptr,
                         byteCnt*8, decryptOut.ptr) != byteCnt*8)
            return 1;

        /* make sure the decrypt output matches original plaintext */
        if (memcmp(plainText.ptr, decryptOut.ptr, byteCnt))
            return 1;

        return 0; /* tests passed! */
    }
    unittest {
        for (int keySize=128; keySize<=256; keySize+=64) {
            for (int testCnt=0; testCnt<10; testCnt++) {
                assert(TestTwofish(MODE_ECB, keySize) == 0,
                       format("ECB Failure at keySize=%d", keySize));
                assert(TestTwofish(MODE_CBC, keySize) == 0,
                       format("CBC Failure at keySize=%d", keySize));
            }
        }
    }
}



